==> cego-2.8.5.tar.gz: cego-2.8.5/src/CegoTableCursor.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoTableCursor.cc
// ------------------
// Cego table cursor class implementation
//      
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoTableCursor
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

// cego includes
#include "CegoTableCursor.h"

#include <string.h>
#include <stdlib.h>

CegoTableCursor::CegoTableCursor()
{
    _pIC = 0;
}

CegoTableCursor::CegoTableCursor(CegoTableManager* pTM, int tabSetId, const Chain& tableName, bool ignoreTouched)
{
    _pTM = pTM;
    _tableName = tableName;
    _pIC = 0;
    _pC = 0;
    _isQualified=false;
    _tabSetId = tabSetId;
    _ignoreTouched = ignoreTouched;
    _idxCached = false;
    _pOE = 0;
}

CegoTableCursor::~CegoTableCursor()
{
    if (_pIC)
    {
	_pIC->abort();
	delete _pIC;
    }
    if ( _pC )
    {
	_pC->abort();
	delete _pC;
    }
}

bool CegoTableCursor::setup( const CegoAttrCond& attrCond)
{

    // cout << "Setting up tablecursor with " << attrCond << endl;

    _attrCond = attrCond;

    if ( _idxCached == false || (Chain)_attrCond.getAttrName() != _idxAttrName )
    {
	_pTM->getObjectListByTable(_tabSetId, _tableName, _idxList, _keyList, _checkList);

	_idxCached = true;
    
	_pOE = _idxList.First();

	_isQualified=false;
	while (_pOE && ! _isQualified)
	{	    
	    if (_pOE->getType() == CegoObject::PINDEX
		|| _pOE->getType() == CegoObject::INDEX
		|| _pOE->getType() == CegoObject::UINDEX )
	    {
		CegoField* pF = _pOE->getSchema().First();
		if (pF)
		{
		    
		    if ((Chain)pF->getAttrName() == (Chain)_attrCond.getAttrName() )
		    {
			// search criteria found
			_idxAttrName = (Chain)_attrCond.getAttrName();
			_idxName = _pOE->getName();
			_idxSchema = _pOE->getSchema();
			_type = _pOE->getType();
			_isQualified = true;
		    }
		}
		if ( _pOE->getType() == CegoObject::UINDEX && ! _isQualified )
		{
		    _idxName = _pOE->getName();
		    _type = _pOE->getType();
		}
	    }
	    _pOE = _idxList.Next();
	}
    }

    return _isQualified;
}

void CegoTableCursor::getIdxSchema(ListT<CegoField>& idxSchema)
{
    idxSchema = _idxSchema;
}

bool CegoTableCursor::getFirst(ListT<CegoField>& fl, CegoDataPointer& dp)
{

    if ( _pTM->getIsolationLevel() == CegoTableManager::READ_UNCOMMITTED )
	_readUncommitted = true;
    else
	_readUncommitted = false;

    if ( _isQualified )
    {
	if (_pIC)
	{
	    _pIC->reset();
	}
	else
	{
	    _pIC = new CegoIndexCursor(_pTM, _tabSetId, _idxName, _type, &_attrCond, _ignoreTouched, _readUncommitted);
	}
	return  _pIC->getFirst(fl, dp);
    }
    else
    {

	if ( _pC )
	{
	    _pC->reset();
	}
	else
	{
	    _pC = _pTM->getObjectCursor(_tabSetId, _tableName, _tableName, CegoObject::TABLE);
	}

	int len;
	char* pc = (char*)_pC->getFirst(len, dp);

	if ( pc && len > 0 )
	{

	    bool useEntry=true;
	    
	    int tid;
	    int tastep;
	    CegoTupleState ts;
	    memcpy(&tid, pc, sizeof(int));

	    char* pctastep = pc + sizeof(int);
	    memcpy(&tastep, pctastep, sizeof(int));

	    char* pcts = pc + sizeof(int) + sizeof(int);
	    memcpy(&ts, pcts, sizeof(CegoTupleState));
	    
	    if ( tid != 0  )
	    {
		if ( _ignoreTouched == true )
		{
		    if ( ts == INSERTED 
			 && tid == _pTM->getTID(_tabSetId) 
			 && tastep < _pTM->getTAStep(_tabSetId) ) 

		    {
			CegoTupleState ts;
			_qh.decodeFVL(fl, pc, len, tid, tastep, ts);
			return true;
		    }
		    else
		    {
			return getNext(fl, dp);
		    }
		}
		else
		{
		    if ( ( _readUncommitted == true 
			   && ts == INSERTED )
			 || ( _readUncommitted == false 
			      && (( ts == INSERTED && tid == _pTM->getTID(_tabSetId)) 
				  || ( ts == DELETED && tid != _pTM->getTID(_tabSetId)))))
		    {
			CegoTupleState ts;
			_qh.decodeFVL(fl, pc, len, tid, tastep, ts);
			return true;
		    }
		    else
		    {
			return getNext(fl, dp);
		    }		
		}
	    }

	    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);

	    return true;
	}
	return false;
    }
}

bool CegoTableCursor::getNext(ListT<CegoField>& fl, CegoDataPointer& dp)
{
       
    if ( _isQualified )
    {
	return _pIC->getNext(fl, dp);
    }
    else
    {

	if ( _pC == 0 )
	{
	    throw Exception(EXLOC, "Invalid ObjectCursor");
	}
	
	do {
	    
	    int len;

	    char* pc = (char*)_pC->getNext(len, dp);

	    if ( pc && len > 0 )
	    {

		bool useEntry=true;
		
		int tid;
		int tastep;

		CegoTupleState ts;

		memcpy(&tid, pc, sizeof(int));

		char* pctastep = pc + sizeof(int);
		memcpy(&tastep, pctastep, sizeof(int));

		char* pcts = pc + sizeof(int) + sizeof(int);
		memcpy(&ts, pcts, sizeof(CegoTupleState));


		if (tid != 0 )
		{
		    
		    if ( _ignoreTouched == true )
		    {
			
			if ( ts == INSERTED 
			     && tid == _pTM->getTID(_tabSetId) 
			     && tastep < _pTM->getTAStep(_tabSetId) )
			{
			    CegoTupleState ts;
			    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);			    
			    return true;   
			}
			else 
			{
			    useEntry = false;
			}
		    }
		    else
		    {
			
			if ( ( _readUncommitted == true 
			       && ts == INSERTED )
			     || ( _readUncommitted == false 
				  && (( ts == INSERTED && tid == _pTM->getTID(_tabSetId)) 
				      || ( ts == DELETED && tid != _pTM->getTID(_tabSetId)))))
			    
			{
			    CegoTupleState ts;
			    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);			    
			    return true;   
			}
			else 
			{
			    useEntry=false;
			}
		    }
		}
		
		if ( useEntry )
		{	       
		    CegoTupleState ts;
		    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);
		    return true;
		}
	    }
	    else
	    {
		return false;
	    }	    
	}  while ( 1 );
    }    
}

const Chain& CegoTableCursor::getIndexName() const
{
    return _idxName;
}

void CegoTableCursor::abort()
{
    if (_pIC)
    {
	_pIC->abort();
    }
    if ( _pC )
    {
	_pC->abort();
    }
}

==> cego-2.12.7.tar.gz: cego-2.12.7/src/CegoTableCursor.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoTableCursor.cc
// ------------------
// Cego table cursor class implementation
//      
// Design and Implementation by Bjoern Lemke
//     
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoTableCursor
// 
// Description: 
//
// Status: QG-2.6
//
///////////////////////////////////////////////////////////////////////////////

// cego includes
#include "CegoTableCursor.h"

#include <string.h>
#include <stdlib.h>

CegoTableCursor::CegoTableCursor()
{
    _pIC = 0;
}

CegoTableCursor::CegoTableCursor(CegoTableManager* pTM, int tabSetId, const Chain& tableName, bool ignoreTouched)
{
    _pTM = pTM;
    _tableName = tableName;
    _pIC = 0;
    _pC = 0;
    _idxMatch = CegoAttrCond::INAPP;
    _tabSetId = tabSetId;
    _ignoreTouched = ignoreTouched;
    _idxSetup = false;
    _pOE = 0;

    _pLogger = _pTM->getDBMng();
    _modId = _pTM->getDBMng()->getModId("CegoTableCursor");
}

CegoTableCursor::~CegoTableCursor()
{
    if (_pIC)
    {
	_pIC->abort();
	delete _pIC;
    }
    if ( _pC )
    {
	_pC->abort();
	delete _pC;
    }
}

CegoAttrCond::IndexMatch CegoTableCursor::setup( const CegoAttrCond& attrCond)
{

    if ( _idxSetup == false || attrCond.diff(_attrCond) == false )
    {
#ifdef DEBUG
	_pLogger->log(_modId, Logger::DEBUG, Chain("Setting up tablecursor with ") + attrCond.toChain());
#endif

	_idxSetup = true;

	_pTM->getObjectListByTable(_tabSetId, _tableName, _idxList, _keyList, _checkList);
    
	_pOE = _idxList.First();
	
	_idxMatch = CegoAttrCond::INAPP;
	
	int strength = 0;
	
	while ( _pOE && _idxMatch != CegoAttrCond::FULL )
	{	    
	    if (_pOE->getType() == CegoObject::PINDEX
		|| _pOE->getType() == CegoObject::INDEX
		|| _pOE->getType() == CegoObject::UINDEX )
	    {

#ifdef DEBUG
		_pLogger->log(_modId, Logger::DEBUG, Chain("Checking index ") + _pOE->getName());
#endif

		CegoAttrCond::IndexMatch indexMatch = attrCond.checkIndex(_pOE->getSchema());

		if ( indexMatch == CegoAttrCond::FULL || indexMatch == CegoAttrCond::PART )
		{

		    CegoAttrCond checkCond = attrCond.getFilterCond(_pOE->getSchema());

		    if ( checkCond.getStrength() > strength )
		    {

			strength = checkCond.getStrength();
#ifdef DEBUG
			_pLogger->log(_modId, Logger::DEBUG, Chain("Strength ") + Chain(strength) + Chain(" tops for ") + checkCond.toChain());
#endif

			_attrCond = checkCond; 
			_attrCond.setIdxSchema(_pOE->getSchema());
			
			_idxName = _pOE->getName();
#ifdef DEBUG
			_pLogger->log(_modId, Logger::DEBUG, Chain("Using index ") + _idxName );
#endif
			
			_idxSchema = _pOE->getSchema();
			_type = _pOE->getType();
			_idxMatch = indexMatch;
		    }
		    else
		    {
#ifdef DEBUG
			_pLogger->log(_modId, Logger::DEBUG, Chain("Strength ") + Chain(strength) + Chain(" to weak for ") + checkCond.toChain());
#endif
		    }
		}
	    }
	    _pOE = _idxList.Next();
	}
    }
    else
    {
#ifdef DEBUG
	_pLogger->log(_modId, Logger::DEBUG, Chain("Tablecursor already set up "));
#endif
    }
    // Setting up attribute condition to actual field values


#ifdef DEBUG    

    if ( _idxMatch == CegoAttrCond::FULL )
    {
	_pLogger->log(_modId, Logger::DEBUG, Chain("Index classified as full qualified"));
    }
    else if ( _idxMatch == CegoAttrCond::PART )
    {
	_pLogger->log(_modId, Logger::DEBUG, Chain("Index classified as part qualified"));
    }
    else
	_pLogger->log(_modId, Logger::DEBUG, Chain("Index classified as not qualified"));
#endif

    return _idxMatch;
}

void CegoTableCursor::getIdxSchema(ListT<CegoField>& idxSchema)
{
    idxSchema = _idxSchema;
}

bool CegoTableCursor::getFirst(ListT<CegoField>& fl, CegoDataPointer& dp)
{
    if ( _pTM->getIsolationLevel() == CegoTableManager::READ_UNCOMMITTED )
	_readUncommitted = true;
    else
	_readUncommitted = false;

    if ( _idxMatch == CegoAttrCond::FULL || _idxMatch == CegoAttrCond::PART  )
    {
	if (_pIC)
	{
	    _pIC->reset();
	}
	else
	{
	    _pIC = new CegoIndexCursor(_pTM, _tabSetId, _idxName, _type, &_attrCond, _ignoreTouched, _readUncommitted);
	}
	return  _pIC->getFirst(fl, dp);
    }
    else
    {

	if ( _pC )
	{
	    _pC->reset();
	}
	else
	{
	    _pC = _pTM->getObjectCursor(_tabSetId, _tableName, _tableName, CegoObject::TABLE);
	}

	int len;
	char* pc = (char*)_pC->getFirst(len, dp);

	if ( pc && len > 0 )
	{

	    bool useEntry=true;
	    
	    int tid;
	    int tastep;
	    CegoTupleState ts;
	    memcpy(&tid, pc, sizeof(int));

	    char* pctastep = pc + sizeof(int);
	    memcpy(&tastep, pctastep, sizeof(int));

	    char* pcts = pc + sizeof(int) + sizeof(int);
	    memcpy(&ts, pcts, sizeof(CegoTupleState));
	    
	    if ( tid != 0  )
	    {
		if ( _ignoreTouched == true )
		{
		    if ( ts == INSERTED 
			 && tid == _pTM->getTID(_tabSetId) 
			 && tastep < _pTM->getTAStep(_tabSetId) ) 

		    {
			CegoTupleState ts;
			_qh.decodeFVL(fl, pc, len, tid, tastep, ts);
			return true;
		    }
		    else
		    {
			return getNext(fl, dp);
		    }
		}
		else
		{
		    if ( ( _readUncommitted == true 
			   && ts == INSERTED )
			 || ( _readUncommitted == false 
			      && (( ts == INSERTED && tid == _pTM->getTID(_tabSetId)) 
				  || ( ts == DELETED && tid != _pTM->getTID(_tabSetId)))))
		    {
			CegoTupleState ts;
			_qh.decodeFVL(fl, pc, len, tid, tastep, ts);
			return true;
		    }
		    else
		    {
			return getNext(fl, dp);
		    }		
		}
	    }


	    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);

	    return true;
	}
	return false;
    }
}

bool CegoTableCursor::getNext(ListT<CegoField>& fl, CegoDataPointer& dp)
{
       
    if ( _idxMatch == CegoAttrCond::FULL || _idxMatch == CegoAttrCond::PART  )
    {
	return _pIC->getNext(fl, dp);
    }
    else
    {

	if ( _pC == 0 )
	{
	    throw Exception(EXLOC, "Invalid ObjectCursor");
	}
	
	do {
	    
	    int len;

	    char* pc = (char*)_pC->getNext(len, dp);

	    if ( pc && len > 0 )
	    {

		bool useEntry=true;
		
		int tid;
		int tastep;

		CegoTupleState ts;

		memcpy(&tid, pc, sizeof(int));

		char* pctastep = pc + sizeof(int);
		memcpy(&tastep, pctastep, sizeof(int));

		char* pcts = pc + sizeof(int) + sizeof(int);
		memcpy(&ts, pcts, sizeof(CegoTupleState));


		if (tid != 0 )
		{
		    
		    if ( _ignoreTouched == true )
		    {
			
			if ( ts == INSERTED 
			     && tid == _pTM->getTID(_tabSetId) 
			     && tastep < _pTM->getTAStep(_tabSetId) )
			{
			    CegoTupleState ts;
			    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);			    
			    return true;   
			}
			else 
			{
			    useEntry = false;
			}
		    }
		    else
		    {
			
			if ( ( _readUncommitted == true 
			       && ts == INSERTED )
			     || ( _readUncommitted == false 
				  && (( ts == INSERTED && tid == _pTM->getTID(_tabSetId)) 
				      || ( ts == DELETED && tid != _pTM->getTID(_tabSetId)))))
			    
			{
			    CegoTupleState ts;
			    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);			    
			    return true;   
			}
			else 
			{
			    useEntry=false;
			}
		    }
		}
		
		if ( useEntry )
		{	       
		    CegoTupleState ts;
		    _qh.decodeFVL(fl, pc, len, tid, tastep, ts);
		    return true;
		}
	    }
	    else
	    {
		return false;
	    }	    
	}  while ( 1 );
    }    
}

const Chain& CegoTableCursor::getIndexName() const
{
    return _idxName;
}

void CegoTableCursor::abort()
{
    if (_pIC)
    {
	_pIC->abort();
    }
    if ( _pC )
    {
	_pC->abort();
    }
}
